#include "super_frame.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../utils/bit_reader.h"

super_frame *new_super_frame(int subchannel_size)
{
  super_frame *ret = calloc(1, sizeof(super_frame));
  if (ret == NULL) abort();

  ret->subchannel_index = subchannel_size / 8;
  ret->audio_super_frame_size = ret->subchannel_index * 110;

  ret->cif_start = -1;

  ret->data = malloc(ret->audio_super_frame_size + ret->subchannel_index * 10);
  if (ret->data == NULL) abort();

  ret->frame_size = (ret->audio_super_frame_size + ret->subchannel_index * 10) / 5;

  return ret;
}

#include <neaacdec.h>

/* code based on ETSI TS 102 563 V2.1.1 */
static void audio_decode_super_frame(super_frame *s)
{
{
  int i;
  printf("superframe[%d]:", s->audio_super_frame_size);
  for (i = 0; i < s->audio_super_frame_size; i++) printf(" %2.2x", s->data[i]);
  printf("\n");
}

  bit_reader b;
  int dac_rate;
  int sbr_flag;
  int aac_channel_mode;
  int ps_flag;
  int mpeg_surround_config;
  int num_aus;
  int n;
  int au_start[6+1];
  int au_size[6];

  init_bit_reader(&b, s->data, s->audio_super_frame_size);

  /* TODO: check/use header_firecode */
  get_bits(&b, 16);                           /* header_firecode */
  get_bits(&b, 1);                            /* rfa */
  dac_rate             = get_bits(&b, 1);
  sbr_flag             = get_bits(&b, 1);
  aac_channel_mode     = get_bits(&b, 1);
  ps_flag              = get_bits(&b, 1);
  mpeg_surround_config = get_bits(&b, 3);

  if (dac_rate == 0 && sbr_flag == 1)      { num_aus = 2; au_start[0] = 5; }
  else if (dac_rate == 1 && sbr_flag == 1) { num_aus = 3; au_start[0] = 6; }
  else if (dac_rate == 0 && sbr_flag == 0) { num_aus = 4; au_start[0] = 8; }
  else                                     { num_aus = 6; au_start[0] = 11; }

  for (n = 1; n < num_aus; n++) {
    au_start[n] = get_bits(&b, 12);
    if (au_start[n] <= au_start[n-1] + 2
       || au_start[n] >= s->audio_super_frame_size - 2) {
      printf("bad superframe, skipping\n");
      return;
    }
  }

  au_start[num_aus] = s->audio_super_frame_size;

  for (n = 0; n < num_aus; n++)
    au_size[n] = au_start[n+1] - au_start[n] - 2;

for (n = 0; n < num_aus; n++) {
  printf("au start %d size %d\n", au_start[n], au_size[n]);
}

  static NeAACDecHandle h;
static int inited = 0;
if (!inited) {
  inited = 1;
  h = NeAACDecOpen();

  unsigned char cf[2];
  cf[0] = (0x02 << 3) | (6 >> 1);
  cf[1] = ((6 & 1) << 7) | (2 << 3) | 0x4;
  unsigned long sample_rate;
  unsigned char channels;
  int ret = NeAACDecInit2(h, cf, 2, &sample_rate, &channels);
}

  for (n = 0; n < num_aus; n++) {
    NeAACDecFrameInfo hinfo;
    short *out = NeAACDecDecode(h, &hinfo, s->data + au_start[n], au_size[n]);
printf("%p\n", out);
static FILE *f = NULL;
if (f == NULL) f = fopen("/tmp/out.raw", "w");
if (f == NULL) abort();
fwrite(out, hinfo.samples, 2, f); fflush(f);
  }
}

void audio_receive_frame(super_frame *s, int cif_id, 
                         unsigned char *data, int len)
{
  if (len != s->frame_size) {
    printf("%s:%d:%s: fatal, bad size\n", __FILE__, __LINE__, __FUNCTION__);
    exit(1);
  }

  /* check continuity of cif_id */
  if (cif_id != s->next_expected_cif) {
    /* we need to resync */
    s->cif_start = -1;
  }

  s->next_expected_cif = (cif_id + 1) % 5000;

  /* resync in progress? */
  if (s->cif_start == -1) {
    /* do real sync, the data I collected seems to do it like this */
    if (cif_id % 5)
      return;

    /* sync done */
    s->cif_start = cif_id;
    s->pos = 0;
  }

  memcpy(s->data + s->pos * s->frame_size, data, len);
  s->pos++;

  /* check full super frame received */
  if (s->pos == 5) {
    audio_decode_super_frame(s);
    s->pos = 0;
  }
}
