#include "msc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../coding/time_interleaving.h"
#include "../coding/puncture.h"
#include "../coding/viterbi_decode.h"
#include "../audio/super_frame.h"

msc_config *new_msc_config(void)
{
  msc_config *ret = calloc(1, sizeof(msc_config));

  ret->cif_id = -1;

  ret->energy_dispersal = new_energy_dispersal(10000);

  return ret;
}

static void reset_msc_data(msc_config *m, int cif_id)
{
  m->in_start = 0;
  m->in_size = 0;
  m->cif_id = cif_id;
  m->in_start_cif_id = m->cif_id;
}

static void decode_subchannel(msc_config *m, sub_channel *c)
{
  int bits[864*64];
  char data[10000];
  int polys[4] = { 0133, 0171, 0145, 0133 };
  int *in;
  int *out;
  int i;

  /* unpuncture */
  //int I = 192*11;
  //int L = 6 * 11;
  int L1 = 6 * 11 - 3; //2 * 11 - 3;
  int L2 = 3; //4 * 11 + 3;
  int PI1 = 8; //14;
  int PI2 = 7; //13;

  in = &m->in[m->in_start][c->start * 64];
  out = bits;

  for (i = 0; i < L1 * 4; i++) {
    unpuncture(PI1, in, out);
    in += PI1 + 8;
    out += 32;
  }

  for (i = 0; i < L2 * 4; i++) {
    unpuncture(PI2, in, out);
    in += PI2 + 8;
    out += 32;
  }

  /* last 24 bits, puncturing patter: 1100 1100 1100 1100 1100 1100 */
  for (i = 0; i < 6; i++) {
    *out++ = *in++;
    *out++ = *in++;
    *out++ = -128;
    *out++ = -128;
  }

printf("yoyoyo out bits %d in bits %d\n", (int)(out-bits), (int)(in-&m->in[m->in_start][c->start * 64]));

  memset(data, 0, sizeof(data));
  viterbi_decode(data, bits, 8472, 7, polys, 4);

  do_energy_dispersal_n(m->energy_dispersal, data, 8472/4/8);

printf("decoded:");
for (i = 0; i<8472/4/8 + 1; i++) printf(" %2.2x", (unsigned char)data[i]);
printf("\n");

  if (m->audio == NULL) m->audio = new_super_frame(88);
  audio_receive_frame(m->audio, m->in_start_cif_id, (unsigned char *)data, (8472-24)/4/8);
}

static void msc_decode_subchannels(msc_config *m)
{
  int i;

  for (i = 0; i < 64; i++) {
    sub_channel *c = &m->sc[i];
    if (!c->configured)
      continue;
if (i == 11)
    decode_subchannel(m, c);
  }
}

void receive_msc_bits(msc_config *m, int cif_id, int *bits, int cif_symbol)
{
printf("receive_msc_bits cif_id %d cif_symbol %d\n", cif_id, cif_symbol);
  int i, j;
  int r;

  if (cif_id != m->cif_id)
    reset_msc_data(m, cif_id);

  /* put data - do time de-interleaving */
  r = (m->in_start + (m->cif_id - m->in_start_cif_id + 5000) % 5000) % 16;

  /* time interleaving is a per-subchannel thing, but the start of a
   * subchannel is multiple of 64 bits (so of 16 too), so we can do it
   * globally here (time interleaving works by blocks of 16 bits)
   */
  for (i = cif_symbol * 3072, j = 0; i < cif_symbol * 3072 + 3072; i++, j++)
    m->in[TIME_DEINTERLEAVE(r, i)][i] = bits[j];

  if (cif_symbol != 17)
    return;

  m->cif_id = (m->cif_id + 1) % 5000;

  /* all symbols for the CIF received, let's process (if we can) */
  m->in_size++;
  if (m->in_size != 16)
    return;

  /* we can decode data for time r-15 (which is time m->in_start) */
  msc_decode_subchannels(m);

  /* get ready for next CIF */
  m->in_start = (m->in_start + 1) % 16;
  m->in_start_cif_id = (m->in_start_cif_id + 1) % 5000;
  m->in_size--;
}

static void check_or_set(sub_channel *new, sub_channel *c)
{
  if (c->configured == 0) {
    *c = *new;
    return;
  }

  if (memcmp(c, new, sizeof(sub_channel))) {
    printf("error: sub channel configuration changed, not handled\n");
    exit(1);
  }
}

void msc_update_config(msc_config *c, frame_figs *f)
{
  int i, j;
  sub_channel sc;

  for (i = 0; i < f->size; i++) {
    if (f->f[i].type == FIG0_1) {
      fig0_1 *ff = &f->f[i].u.fig0_1;
      for (j = 0; j < ff->n; j++) {
        if (ff->sub_channel[j].is_long && ff->sub_channel[j].option == 0) {
          int scdiv[4] = { 12, 8, 6, 4 };
          int n = ff->sub_channel[j].sc_size / scdiv[ff->sub_channel[j].protection_level];
          if (ff->sub_channel[j].sc_size % scdiv[ff->sub_channel[j].protection_level]) {
            printf("error: bad size in FIG 0/1, skipping sub channel\n");
            continue;
          }
          sc.configured = 1;
          sc.start = ff->sub_channel[j].start_address;
          switch (ff->sub_channel[j].protection_level) {
          case 0 /* 1-A */:
            sc.l[0]  = 6 * n - 3; sc.l[1]  = 3;
            sc.pi[0] = 24;        sc.pi[1] = 23;
            break;
          case 1 /* 2-A */:
            sc.l[0]  = 2 * n - 3; sc.l[1]  = 4 * n + 3;
            sc.pi[0] = 14;        sc.pi[1] = 13;
            break;
          case 2 /* 3-A */:
            sc.l[0]  = 6 * n - 3; sc.l[1]  = 3;
            sc.pi[0] = 8;         sc.pi[1] = 7;
            break;
          case 3 /* 4-A */:
            sc.l[0]  = 4 * n - 3; sc.l[1]  = 2 * n + 3;
            sc.pi[0] = 3;         sc.pi[1] = 2;
            break;
          }
          if (ff->sub_channel[j].protection_level == 1 /* 2-A */ && n == 1) {
            sc.l[0]  = 5;  sc.l[1] = 1;
            sc.pi[0] = 13; sc.pi[1] = 12;
          }
          sc.l[2] = sc.l[3] = 0;
          sc.pi[2] = sc.pi[3] = 0;
          check_or_set(&sc, &c->sc[ff->sub_channel[j].sc_id]);
        } else {
          printf("%s:%d:%s: TODO\n", __FILE__, __LINE__, __FUNCTION__);
          exit(1);
        }
      }
    }
  }
}
